from __future__ import absolute_import, division, print_function

__all__ = [
    "__title__", "__summary__", "__version__", "__author__", "__email__", "__license__", "__copyright__", "__url__"
]

__title__ = "constant_sorrow"

__url__ = "https://github.com/nucypher/constantSorrow"

__summary__ = "I am the man of constant sorrow; I've seen special values as ints, all my days."

__version__ = "v0.1.0-alpha.7"

__author__ = "NuCypher"

__license__ = "Apache Software License v2.0"

__email__ = "dev@nucypher.com"

__copyright__ = 'Copyright (C) 2018 NuCypher'
